package nachos.threads;

import nachos.machine.*;

/**
 * A <i>communicator</i> allows threads to synchronously exchange 32-bit
 * messages. Multiple threads can be waiting to <i>speak</i>,
 * and multiple threads can be waiting to <i>listen</i>. But there should never
 * be a time when both a speaker and a listener are waiting, because the two
 * threads can be paired off at this point.
 */
public class Communicator {
    /**
     * A condition used to synchronize interactions between the speakers and
     * listeners.
     */
    private final Condition condition;
    
    /** A lock used by the condition. */
    private final Lock lock = new Lock();
    
    /**
     * A boxed word sent from a speaker to a listener.  When this value is null,
     * a speaker can write to it.  When this value is not null, a listener may
     * read it.
     */
    private Integer word;

    /**
     * Allocate a new communicator.
     */
    public Communicator() {
        this.condition = new Condition(this.lock);
    }

    /**
     * A helper method that returns true when a word is available for a
     * listener.  This method is accessed only when the lock is acquired.
     *
     * @return True indicates a word is available for a listener.
     */
    private boolean canListen() {
        return this.word != null;
    }

    /**
     * A helper method that returns true when a speaker may store a word for a
     * listener.  This method is accessed only when the lock is acquired.
     *
     * @return True indicates the speaker may store a word for a listener.
     */
    private boolean canSpeak() {
        return this.word == null;
    }

    /**
     * Wait for a thread to listen through this communicator, and then transfer
     * <i>word</i> to the listener.
     *
     * <p>
     * Does not return until this thread is paired up with a listening thread.
     * Exactly one listener should receive <i>word</i>.
     *
     * @param	word	the integer to transfer.
     */
    public void speak(int word) {
        lock.acquire();

        // Wait until the speaker can write the word.
        while (!this.canSpeak())
            condition.sleep();

        // Write the word and inform any listeners it is safe to read again.
        this.word = new Integer(word);
        condition.wakeAll();

        // Wait for a listener to receive the word before returning.
        while (this.canListen())
            condition.sleep();

        lock.release();
    }

    /**
     * Wait for a thread to speak through this communicator, and then return
     * the <i>word</i> that thread passed to <tt>speak()</tt>.
     *
     * @return	the integer transferred.
     */    
    public int listen() {
        lock.acquire();

        // Wait until a speaker has written a word.
        while (!this.canListen())
            condition.sleep();

        // Retreive the word and inform any speakers it is safe to write again.
        final int result = this.word.intValue();
        this.word = null;
        condition.wakeAll();

        lock.release();
        return result;
    }

    /**
     * Tests whether this module is working.
     */
    public static void selfTest() {
        CommunicatorTest.runTest();
    }
}
